///////////////////////////////////////////////////////////////////////////////
// This source file is part of the LuaPlus source distribution and is Copyright
// 2002 by Joshua C. Jensen (jjensen@workspacewhiz.com).
//
// The latest version may be obtained from http://wwhiz.com/LuaPlus/.
//
// The code presented in this file may be freely used and modified for all
// non-commercial and commercial purposes.
///////////////////////////////////////////////////////////////////////////////
#pragma once

///////////////////////////////////////////////////////////////////////////////
// namespace LuaPlus
///////////////////////////////////////////////////////////////////////////////
namespace LuaPlus
{

/**
**/
namespace LuaHelper
{
	/**
		Attempts retrieval of the value from the passed in LuaStackObject.

		\param obj The LuaStackObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaStackObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static bool GetBoolean( LuaStackObject& obj, KeyT key, bool require = true, bool defaultValue = false )
	{
		LuaAutoObject boolObj = obj[ key ];
		if ( !boolObj.IsBoolean() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return boolObj.GetBoolean();
	}


	/**
		Attempts retrieval of the value from the passed in LuaStackObject.

		\param obj The LuaStackObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaStackObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static int GetInteger( LuaStackObject& obj, KeyT key, bool require = true, int defaultValue = -1 )
	{
		LuaAutoObject intObj = obj[ key ];
		if ( !intObj.IsNumber() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return intObj.GetInteger();
	}


	/**
		Attempts retrieval of the value from the passed in LuaStackObject.

		\param obj The LuaStackObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaStackObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static float GetFloat( LuaStackObject& obj, KeyT key, bool require = true, float defaultValue = -1.0f )
	{
		LuaAutoObject floatObj = obj[ key ];
		if ( !floatObj.IsNumber() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return floatObj.GetNumber();
	}


	/**
		Attempts retrieval of the value from the passed in LuaStackObject.

		\param obj The LuaStackObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaStackObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static void* GetLightUserData( LuaStackObject& obj, KeyT key, bool require = true, void* defaultValue = NULL )
	{
		LuaAutoObject outObj = obj[ key ];
		if ( !outObj.IsLightUserData() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return outObj.GetLightUserData();
	}


	/**
		Attempts retrieval of the value from the passed in LuaStackObject.

		\param obj The LuaStackObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaStackObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static const char* GetString( LuaStackObject& obj, KeyT key, bool require = true, const char* defaultValue = "" )
	{
		LuaAutoObject stringObj = obj[ key ];
		if ( !stringObj.IsString() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return stringObj.GetString();
	}


	/**
		Attempts retrieval of the obj from the passed in LuaStackObject.

		\param obj The LuaStackObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaStackObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\return Returns the object found.
	**/
	template< typename KeyT >
	static LuaStackObject GetTable( LuaStackObject& obj, KeyT key, bool require = true )
	{
		LuaStackObject tableObj = obj[ key ];
		if ( !tableObj.IsTable() )
		{
			if ( require )
				assert( 0 );
		}
		return tableObj;
	}

	/**
		Attempts retrieval of the value from the passed in LuaObject.

		\param obj The LuaObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static bool GetBoolean( LuaObject& obj, KeyT key, bool require = true, bool defaultValue = false )
	{
		LuaObject boolObj = obj[ key ];
		if ( !boolObj.IsBoolean() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return boolObj.GetBoolean();
	}


	/**
		Attempts retrieval of the value from the passed in LuaObject.

		\param obj The LuaObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static int GetInteger( LuaObject& obj, KeyT key, bool require = true, int defaultValue = -1 )
	{
		LuaObject intObj = obj[ key ];
		if ( !intObj.IsNumber() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return intObj.GetInteger();
	}


	/**
		Attempts retrieval of the value from the passed in LuaObject.

		\param obj The LuaObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static float GetFloat( LuaObject& obj, KeyT key, bool require = true, float defaultValue = -1.0f )
	{
		LuaObject floatObj = obj[ key ];
		if ( !floatObj.IsNumber() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return floatObj.GetNumber();
	}


	/**
		Attempts retrieval of the value from the passed in LuaObject.

		\param obj The LuaObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static void* GetLightUserData( LuaObject& obj, KeyT key, bool require = true, void* defaultValue = NULL )
	{
		LuaObject outObj = obj[ key ];
		if ( !outObj.IsLightUserData() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return outObj.GetLightUserData();
	}


	/**
		Attempts retrieval of the value from the passed in LuaObject.

		\param obj The LuaObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\param defaultValue The default value to be returned if require is
			not true and the key doesn't exist or the found value is not
			of the right type.
		\return Returns the value found or the defaultValue.
	**/
	template< typename KeyT >
	static const char* GetString( LuaObject& obj, KeyT key, bool require = true, const char* defaultValue = "" )
	{
		LuaObject stringObj = obj[ key ];
		if ( !stringObj.IsString() )
		{
			if ( require )
				assert( 0 );
			return defaultValue;
		}
		return stringObj.GetString();
	}


	/**
		Attempts retrieval of the obj from the passed in LuaObject.

		\param obj The LuaObject representing a table to attempt retrieval
			of the value from.
		\param key The lookup key for the value.  The key may be of number,
			string, or LuaObject type.
		\param require If set to true, the key must exist in the table and
			must be of the requested type.  If it doesn't exist or isn't of
			the type, an assertion fires.  Otherwise, defaultValue is 
			returned.
		\return Returns the object found.
	**/
	template< typename KeyT >
	static LuaObject GetTable( LuaObject& obj, KeyT key, bool require = true )
	{
		LuaObject tableObj = obj[ key ];
		if ( !tableObj.IsTable() )
		{
			if ( require )
				assert( 0 );
		}
		return tableObj;
	}
};

} // namespace LuaPlus


