///////////////////////////////////////////////////////////////////////////////
// This source file is part of the LuaPlus source distribution and is Copyright
// 2002 by Joshua C. Jensen (jjensen@workspacewhiz.com).
//
// The latest version may be obtained from http://wwhiz.com/LuaPlus/.
//
// The code presented in this file may be freely used and modified for all
// non-commercial and commercial purposes.
///////////////////////////////////////////////////////////////////////////////
#define BUILDING_LUAPLUS
#include "LuaLink.h"
LUA_EXTERN_C_BEGIN
#include "lfunc.h"
#include "lobject.h"
#include "lstate.h"
#include "lstring.h"
#include "ltable.h"
#include "lvm.h"
LUA_EXTERN_C_END
#include "LuaPlus.h"
#include "LuaPlusInternals.h"
#include <new.h>

NAMESPACE_LUA_BEGIN
LUA_EXTERN_C TObject *negindex (lua_State *L, int index);
NAMESPACE_LUA_END

USING_NAMESPACE_LUA

///////////////////////////////////////////////////////////////////////////////
// namespace LuaPlus
///////////////////////////////////////////////////////////////////////////////
namespace LuaPlus
{

void lua_pushtobject(lua_State *L, void* tobject);

#ifndef api_check
#define api_check(L, o)		(assert(o))
#endif

#define luaA_index(L, index) \
  ( (index > 0) ? \
    (api_check(L, index <= L->top - L->ci->base), L->ci->base+index-1) : \
    negindex(L, index))


/**
	Various constructors accepting different parameters.
**/
LuaObject::LuaObject(LuaState* state) :
	m_state(state)
{
	AddToUsedList(this);
	setnilvalue(&m_object);
}


/**
	Various constructors accepting different parameters.
**/
LuaObject::LuaObject( LuaState* state, int stackIndex ) :
	m_state(state)
{
	AddToUsedList(this);
	StkId t;
	t = luaA_index(m_state->m_state, stackIndex);
	m_object = *t;
}

/**
	Various constructors accepting different parameters.
**/
LuaObject::LuaObject( LuaState& state, int stackIndex ) :
	m_state(&state)
{
	AddToUsedList(this);
	m_object = *luaA_index(m_state->m_state, stackIndex);
}

/**
	Various constructors accepting different parameters.
**/
inline LuaObject::LuaObject( LuaState* state, const TObject* obj ) :
	m_state(state)
{
	AddToUsedList(this);
	m_object = *obj;
}

/**
	Copy constructor.
**/
LuaObject::LuaObject( const LuaStackObject& src ) :
	m_state(src.m_state)
{
	AddToUsedList(this);
	m_object = *luaA_index(m_state->m_state, src.m_stackIndex);
}

/**
	Copy constructor.
**/
LuaObject::LuaObject( const LuaObject& src ) :
	m_state( src.m_state )
{
	AddToUsedList(this);
	m_object = src.m_object;
}

/**
	Assignment operator.
**/
const LuaObject& LuaObject::operator=( const LuaObject& src )
{
	if (m_state)
		RemoveFromUsedList(this);
	m_state = src.m_state;
	AddToUsedList(this);
	m_object = src.m_object;
	return *this;
}

/**
	Assignment operator.
**/
const LuaObject& LuaObject::operator=( const LuaStackObject& src )
{
	if (m_state)
		RemoveFromUsedList(this);
	m_state = src.m_state;
	AddToUsedList(this);
	m_object = *luaA_index(m_state->m_state, src.m_stackIndex);
	return *this;
}

LuaObject::~LuaObject()
{
	if (m_state)
		RemoveFromUsedList(this);
}


inline void LuaObject::AddToUsedList(LuaObject* object)
{
	LuaObject& headObject = *m_state->GetHeadObject();
	object->m_next = headObject.m_next;
	headObject.m_next = object;
	object->m_next->m_prev = object;
	object->m_prev = &headObject;
}

inline void LuaObject::RemoveFromUsedList(LuaObject* object)
{
	// remove pOldNode from list
	object->m_prev->m_next = object->m_next;
	object->m_next->m_prev = object->m_prev;
}

/**
	Retrieves the LuaState object associated with this LuaStackObject.
**/
const char* LuaObject::GetTypeName() const
{
	assert(m_state);
	const char* typeName = lua_typename(m_state->m_state, GetType());
	return typeName;
}

int LuaObject::GetType() const
{
	assert(m_state);
	return ttype(&m_object);
}

bool LuaObject::IsNil() const
{
	assert(m_state);
	return ttisnil(&m_object);
}

bool LuaObject::IsTable() const
{
	assert(m_state);
	return ttistable(&m_object);
}

bool LuaObject::IsUserData() const
{
	assert(m_state);
	return ttisuserdata(&m_object)  ||  ttislightuserdata(&m_object);
}

bool LuaObject::IsCFunction() const
{
	assert(m_state);
	return ttisfunction(&m_object)  &&  clvalue(&m_object)->c.isC;
}

bool LuaObject::IsNumber() const
{
	assert(m_state);
	return ttisnumber(&m_object);
}

bool LuaObject::IsString() const
{
	assert(m_state);
	return ttisstring(&m_object);
}

bool LuaObject::IsWString() const
{
	assert(m_state);
	return ttiswstring(&m_object);
}

bool LuaObject::IsFunction() const
{
	assert(m_state);
	return ttisfunction(&m_object);
}

bool LuaObject::IsNone() const
{
	assert(m_state);
	return ttype(&m_object) == LUA_TNONE;
}

bool LuaObject::IsLightUserData() const
{
	assert(m_state);
	return ttislightuserdata(&m_object);
}

bool LuaObject::IsBoolean() const
{
	assert(m_state);
	return ttisboolean(&m_object);
}

int LuaObject::GetInteger() const
{
	assert(m_state);
	assert(IsNumber());
	return (int)nvalue(&m_object);
}

float LuaObject::GetFloat() const
{
	assert(m_state);
	assert(IsNumber());
	return (float)nvalue(&m_object);
}

double LuaObject::GetDouble() const
{
	assert(m_state);
	assert(IsNumber());
	return (double)nvalue(&m_object);
}

float LuaObject::GetNumber() const
{
	assert(m_state);
	assert(IsNumber());
	return (float)nvalue(&m_object);
}

const char* LuaObject::GetString() const
{
	assert(m_state);
	assert(IsString());
	return svalue(&m_object);
}

const wchar_t* LuaObject::GetWString()	const
{
	assert(m_state);
	assert(IsWString());
	return wsvalue(&m_object);
}

int LuaObject::StrLen()
{
	assert(m_state);
	StkId o = &m_object;
	if (ttisstring(o)  ||  ttiswstring(o))
		return tsvalue(o)->tsv.len;
	assert(0);		// Wasn't a string.
	return 0;
}

NAMESPACE_LUA_PREFIX lua_CFunction LuaObject::GetCFunction() const
{
	assert(m_state);
	assert(IsCFunction());
	return (!IsCFunction()) ? NULL : clvalue(&m_object)->c.f;
}

void* LuaObject::GetUserData()
{
	assert(m_state);
	assert(IsUserData());
	StkId o = &m_object;
	switch (ttype(o))
	{
		case LUA_TUSERDATA:
		{
			if (uvalue(o)->uv.len & 4)
				return *(void**)(uvalue(o) + 1);
			else
				return (uvalue(o) + 1);
		}
		case LUA_TLIGHTUSERDATA: return pvalue(o);
		default: return NULL;
	}
}

const void* LuaObject::GetLuaPointer()
{
	assert(m_state);
	StkId o = &m_object;
	switch (ttype(o))
	{
		case LUA_TTABLE: return hvalue(o);
		case LUA_TFUNCTION: return clvalue(o);
		case LUA_TUSERDATA:
		case LUA_TLIGHTUSERDATA:
			return GetUserData();
		default: return NULL;
	}
}

void* LuaObject::GetLightUserData() const
{
	assert(m_state);
	assert(IsLightUserData());
	return pvalue(&m_object);
}

bool LuaObject::GetBoolean() const
{
	assert(m_state);
	assert(IsBoolean()  ||  IsNil());
	return !l_isfalse(&m_object);
}


/**
**/
LuaObject LuaObject::Clone()
{
	if (IsTable())
	{
		LuaObject tableObj(m_state);
		sethvalue(&tableObj.m_object, luaH_new(m_state->m_state, hvalue(&m_object)->sizearray, hvalue(&m_object)->lsizenode));

		for (LuaTableIterator it(*this); it; ++it)
		{
			if (it.GetValue().IsTable())
			{
				LuaObject clonedChildTableObj = it.GetValue().Clone();
				tableObj.SetObject(it.GetKey(), clonedChildTableObj);
			}
			else
				tableObj.SetObject(it.GetKey(), it.GetValue());
		}

		return tableObj;
	}

	return LuaObject(m_state, &m_object);
}


LuaStackObject LuaObject::PushStack()
{
	assert(m_state);
	lua_pushtobject(m_state->m_state, &m_object);
	return LuaStackObject(m_state, m_state->GetTop());
}


///////////////////////////////////////////////////////////////////////////////
LuaStackObject LuaObject::PCall(int nresults, const char* argTypes, va_list args)
{
	assert(m_state);
	assert(IsFunction());
	LuaStackObject obj = m_state->GetGlobals_Stack()["_ALERT"];
	int startResults = m_state->GetTop() + 1;
	PushStack();

	int numArgs = 0;
	const char* p = argTypes;
	while (*p != '\0')
	{
		switch (*p++)
		{
			case 'f':
				m_state->PushNumber(va_arg(args, float));
				break;
			case 'd':
				m_state->PushNumber(va_arg(args, double));
				break;
			case 'i':
				m_state->PushNumber(va_arg(args, int));
				break;
			case 's':
				m_state->PushString(va_arg(args, char*));
				break;
			case 'n':
				m_state->PushNil();
				break;
			case 'c':
				m_state->PushCClosure(va_arg(args, LuaStateCFunction), 0);
				break;
			case 'b':
				m_state->PushBoolean(va_arg(args, bool));
				break;
			case 'U':
//				m_state->PushUserData(va_arg(args, void*));
				break;
			case 'u':
				m_state->PushLightUserData(va_arg(args, void*));
				break;
			case 'o':
			{
				LuaObject* obj = va_arg(args, LuaObject*);
				obj->PushStack();
				break;
			}
			case 'O':
			{
				LuaStackObject obj = va_arg(args, LuaStackObject);
				obj.Push();
				break;
			}
			default:
				assert(0);
		}
		numArgs++;
	}

	int err = m_state->PCall(numArgs, nresults, obj);	err;
	assert(err == 0);
	obj.Pop();
	return LuaStackObject(m_state, startResults - 1);
}


LuaStackObject LuaObject::PCall(const char* argTypes, ...)
{
	assert(m_state);
	va_list args;
	va_start(args, argTypes);

	LuaStackObject resultsObj = PCall(LUA_MULTRET, argTypes, args);

	va_end(args);

	return resultsObj;
}


LuaStackObject LuaObject::PCall(int nresults, const char* argTypes, ...)
{
	assert(m_state);
	va_list args;
	va_start(args, argTypes);

	LuaStackObject resultsObj = PCall(nresults, argTypes, args);

	va_end(args);

	return resultsObj;
}


LuaObject LuaObject::GetMetaTable()
{
	assert(m_state);
	Table *mt = luaT_getmetatable(m_state->m_state, &m_object);
	TObject valueObj;
	sethvalue(&valueObj, mt);
	return LuaObject(m_state, &valueObj);
}


void LuaObject::SetMetaTable( LuaObject& valueObj )
{
	assert(m_state);
	luaT_setmetatable(m_state->m_state, &m_object, hvalue(&valueObj.m_object));
}


int LuaObject::GetN()
{
	assert(m_state);
	LuaAutoBlock autoBlock(*this);
	PushStack();
	LuaObject tableObj = m_state->GetGlobal("table");
	LuaObject funcObj = tableObj["getn"];
	assert(funcObj.IsFunction());
	LuaStackObject resultObj = funcObj() << *this << LuaRun();
	return resultObj.GetInteger();
}


void LuaObject::SetN(int n)
{
	assert(m_state);
	LuaAutoBlock autoBlock(*this);
	PushStack();
	LuaObject tableObj = m_state->GetGlobal("table");
	LuaObject funcObj = tableObj["setn"];
	assert(funcObj.IsFunction());
	funcObj() << *this << n << LuaRun();
}


void LuaObject::Insert(int index, LuaObject& obj)
{
	assert(m_state);
	LuaAutoBlock autoBlock(*this);
	LuaObject tableObj = m_state->GetGlobal("table");
	LuaObject funcObj = tableObj["insert"];
	assert(funcObj.IsFunction());
	funcObj() << *this << index << obj << LuaRun();
}

	
void LuaObject::Remove(int index)
{
	assert(m_state);
	LuaAutoBlock autoBlock(*this);
	LuaObject tableObj = m_state->GetGlobal("table");
	LuaObject funcObj = tableObj["remove"];
	assert(funcObj.IsFunction());
	funcObj() << *this << index << LuaRun();
}

	
void LuaObject::Sort()
{
	assert(m_state);
	LuaAutoBlock autoBlock(*this);
	LuaObject tableObj = m_state->GetGlobal("table");
	LuaObject funcObj = tableObj["sort"];
	assert(funcObj.IsFunction());
	funcObj() << *this << LuaRun();
}

	
int LuaObject::GetCount()
{
	assert(m_state);
	PushStack();
	int count = lua_getn(m_state->m_state, m_state->GetTop());
	m_state->Pop();
	return count;
}


int LuaObject::GetTableCount()
{
	int count = 0;
	for (LuaTableIterator it(*this); it; ++it)
	{
		count++;
	}
	return count;
}


/**
	Creates a table called [name] within the current LuaStackObject.

	@param name The name of the table to create.
	@param size The size of the table.
	@return Returns the object representing the newly created table.
**/
LuaObject LuaObject::CreateTable(const char* key, int narray, int lnhash)
{
	assert(m_state);
	TObject valueObj;
	sethvalue(&valueObj, luaH_new(m_state->m_state, narray, lnhash));
	SetTableHelper(key, valueObj);
	return LuaObject(m_state, &valueObj);
}


/**
	Creates a table called [key] within the current LuaStackObject.

	@param index The index of the table to create.
	@param size The size of the table.
	@return Returns the object representing the newly created table.
**/
LuaObject LuaObject::CreateTable(int key, int narray, int lnhash)
{
	assert(m_state);
	TObject valueObj;
	sethvalue(&valueObj, luaH_new(m_state->m_state, narray, lnhash));
	SetTableHelper(key, valueObj);
	return LuaObject(m_state, &valueObj);
}
	

/**
	Creates a table called [key] within the current LuaStackObject.

	@param index The index of the table to create.
	@param size The size of the table.
	@return Returns the object representing the newly created table.
**/
LuaObject LuaObject::CreateTable(LuaObject& key, int narray, int lnhash)
{
	assert(m_state);
	TObject valueObj;
	sethvalue(&valueObj, luaH_new(m_state->m_state, narray, lnhash));
	SetTableHelper(key, valueObj);
	return LuaObject(m_state, &valueObj);
}
	

LuaObject LuaObject::GetByName( const char* name )
{
	assert(m_state);
	lua_State* L = m_state->m_state;
	TObject str;
	setsvalue(&str, luaS_newlstr(L, name, strlen(name)));
	const TObject* v = luaV_gettable(L, &m_object, &str, 0);
	return LuaObject(m_state, v);
}

LuaObject LuaObject::GetByIndex( int index )
{
	assert(m_state);
	lua_State* L = m_state->m_state;
#ifdef _DEBUG
	StkId o = &m_object;
	api_check(L, ttistable(o));
#endif _DEBUG
	TObject obj;
	setnvalue(&obj, index);
	const TObject* v = luaV_gettable(L, &m_object, &obj, 0);
	return LuaObject(m_state, v);
}

LuaObject LuaObject::GetByObject( const LuaObject& obj )
{
	assert(m_state);
	const TObject* v = luaV_gettable(m_state->m_state, &m_object, (TObject*)&obj.m_object, 0);
	return LuaObject(m_state, v);
}


LuaObject LuaObject::GetByObject( const LuaStackObject& obj )
{
	assert(m_state);
	lua_State* L = *this;
	const TObject* v = luaV_gettable(m_state->m_state, &m_object, luaA_index(L, obj.m_stackIndex), 0);
	return LuaObject(m_state, v);
}


LuaObject LuaObject::operator[]( const char* name )
{
	assert(m_state);
	lua_State* L = m_state->m_state;
	TObject str;
	setsvalue(&str, luaS_newlstr(L, name, strlen(name)));
	const TObject* v = luaH_get(hvalue(&m_object), &str);
	return LuaObject(m_state, v);
}

LuaObject LuaObject::operator[]( int index )
{
	assert(m_state);
	StkId o = &m_object;
	api_check(L, ttistable(o));
	const TObject* v = luaH_getnum(hvalue(o), index);
	return LuaObject(m_state, v);
}

LuaObject LuaObject::operator[]( const LuaObject& obj )
{
	assert(m_state);
	const TObject* v = luaH_get(hvalue(&m_object), &obj.m_object);
	return LuaObject(m_state, v);
}

LuaObject LuaObject::operator[]( const LuaStackObject& obj )
{
	assert(m_state);
	lua_State* L = *this;
	const TObject* v = luaH_get(hvalue(&m_object), luaA_index(L, obj.m_stackIndex));
	return LuaObject(m_state, v);
}

void LuaObject::SetNil( const char* key )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setnilvalue(&valueObj);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetNil( int key )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setnilvalue(&valueObj);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetNil( LuaObject& key )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setnilvalue(&valueObj);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetBoolean( const char* key, bool value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setbvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}

void LuaObject::SetBoolean( int key, bool value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setbvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetBoolean( LuaObject& key, bool value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setbvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetNumber( const char* key, double value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setnvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetNumber( int key, double value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setnvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetNumber( LuaObject& key, double value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setnvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetString( const char* key, const char* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	if (value == NULL)
		setnilvalue(&valueObj);
	else
		setsvalue(&valueObj, luaS_newlstr(m_state->m_state, value, strlen(value)));
	SetTableHelper(key, valueObj);
}

void LuaObject::SetString( int key, const char* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	if (value == NULL)
		setnilvalue(&valueObj);
	else
		setsvalue(&valueObj, luaS_newlstr(m_state->m_state, value, strlen(value)));
	SetTableHelper(key, valueObj);
}


void LuaObject::SetString( LuaObject& key, const char* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	if (value == NULL)
		setnilvalue(&valueObj);
	else
		setsvalue(&valueObj, luaS_newlstr(m_state->m_state, value, strlen(value)));
	SetTableHelper(key, valueObj);
}


void LuaObject::SetWString( const char* key, const wchar_t* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	if (value == NULL)
		setnilvalue(&valueObj);
	else
		setwsvalue(&valueObj, luaS_newlwstr(m_state->m_state, value, wcslen(value)));
	SetTableHelper(key, valueObj);
}


void LuaObject::SetWString( int key, const wchar_t* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	if (value == NULL)
		setnilvalue(&valueObj);
	else
		setwsvalue(&valueObj, luaS_newlwstr(m_state->m_state, value, wcslen(value)));
	SetTableHelper(key, valueObj);
}


void LuaObject::SetWString( LuaObject& key, const wchar_t* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	if (value == NULL)
		setnilvalue(&valueObj);
	else
		setwsvalue(&valueObj, luaS_newlwstr(m_state->m_state, value, wcslen(value)));
	SetTableHelper(key, valueObj);
}


void LuaObject::SetUserData( const char* key, void* value )
{
	assert(m_state);
	assert(IsTable());
	Udata* u = luaS_newudata(m_state->m_state, 4);
	u->uv.len = 4;  // user data box bit is set.
	*(void**)(u + 1) = value;

	TObject valueObj;
	setuvalue(&valueObj, u);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetUserData( int key, void* value )
{
	assert(m_state);
	assert(IsTable());
	Udata* u = luaS_newudata(m_state->m_state, 4);
	u->uv.len = 4;  // user data box bit is set.
	*(void**)(u + 1) = value;

	TObject valueObj;
	setuvalue(&valueObj, u);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetUserData( LuaObject& key, void* value )
{
	assert(m_state);
	assert(IsTable());
	Udata* u = luaS_newudata(m_state->m_state, 4);
	u->uv.len = 4;  // user data box bit is set.
	*(void**)(u + 1) = value;

	TObject valueObj;
	setuvalue(&valueObj, u);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetLightUserData( const char* key, void* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setpvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetLightUserData( int key, void* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setpvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetLightUserData( LuaObject& key, void* value )
{
	assert(m_state);
	assert(IsTable());
	TObject valueObj;
	setpvalue(&valueObj, value);
	SetTableHelper(key, valueObj);
}


void LuaObject::SetObject( const char* key, LuaObject& value )
{
	assert(m_state);
	assert(IsTable());
	SetTableHelper(key, value.m_object);
}


void LuaObject::SetObject( int key, LuaObject& value )
{
	assert(m_state);
	assert(IsTable());
	SetTableHelper(key, value.m_object);
}


void LuaObject::SetObject( LuaObject& key, LuaObject& value )
{
	assert(m_state);
	assert(IsTable());
	SetTableHelper(key, value.m_object);
}


void LuaObject::AssignNil()
{
	assert(m_state);
	setnilvalue(&m_object);
}

	
void LuaObject::AssignBoolean(bool value)
{
	assert(m_state);
	setbvalue(&m_object, value);
}

	
void LuaObject::AssignNumber(double value)
{
	assert(m_state);
	setnvalue(&m_object, value);
}

	
void LuaObject::AssignString(const char* value)
{
	assert(m_state);
	if (value == NULL)
		setnilvalue(&m_object);
	else
		setsvalue(&m_object, luaS_newlstr(m_state->m_state, value, strlen(value)));
}

	
void LuaObject::AssignWString(const wchar_t* value)
{
	assert(m_state);
	if (value == NULL)
		setnilvalue(&m_object);
	else
		setwsvalue(&m_object, luaS_newlwstr(m_state->m_state, value, wcslen(value)));
}

	
void LuaObject::AssignUserData(void* value)
{
	assert(m_state);
	Udata* u = luaS_newudata(m_state->m_state, 4);
	u->uv.len = 4;  // user data box bit is set.
	*(void**)(u + 1) = value;
	setuvalue(&m_object, u);
}

	
void LuaObject::AssignLightUserData(void* value)
{
	assert(m_state);
	setpvalue(&m_object, value);
}

	
void LuaObject::AssignObject(LuaObject& value)
{
	assert(m_state);
	m_object = value.m_object;
}

	
void LuaObject::AssignNewTable(int narray, int numhash)
{
	assert(m_state);
	int lnhash = luaO_log2(numhash) + 1;
	sethvalue(&m_object, luaH_new(m_state->m_state, narray, lnhash));
}

	
void LuaObject::AssignTObject(TObject* value)
{
	assert(m_state);
	m_object = *value;
}

	
/**
	Assuming the current object is a table, registers a C function called 
	[funcName] within the table.

	@param funcName The name of the function to register.
	@param function A pointer to the C function to register.
**/
void LuaObject::Register( const char* funcName, NAMESPACE_LUA_PREFIX lua_CFunction function )
{
	assert(m_state);
	Closure* cl = luaF_newCclosure(m_state->m_state, 0);
	cl->c.functiontype = 0;
	cl->c.f = function;

	TObject valueObj;
	setclvalue(&valueObj, cl);
	SetTableHelper(funcName, valueObj);
}


/**
	Assuming the current object is a table, registers a C function called 
	[funcName] within the table.

	@param funcName The name of the function to register.
	@param function A pointer to the C function to register.
**/
void LuaObject::Register( const char* funcName, LuaStateCFunction function )
{
	assert(m_state);
	Closure* cl = luaF_newCclosure(m_state->m_state, 0);
	cl->c.functiontype = 2;
	CClosureEx* closureEx = (CClosureEx*)(&cl->c + 1);
	closureEx->f2 = function;

	TObject valueObj;
	setclvalue(&valueObj, cl);
	SetTableHelper(funcName, valueObj);
}

/**
	Assuming the current object is a table, registers a C function called 
	[funcName] within the table.

	@param funcName The name of the function to register.
	@param function A pointer to the C function to register.
**/
void LuaObject::Register( const char* funcName, LSFunctor functor )
{
	assert(m_state);
	Closure* cl = luaF_newCclosure(m_state->m_state, 0);
	cl->c.f = NULL;
	cl->c.functiontype = 1;
	CClosureEx* closureEx = (CClosureEx*)(&cl->c + 1);
	::new(&closureEx->f1) LSFunctor;
	LSFunctor* _functor = (LSFunctor*)&closureEx->f1;
	*_functor = functor;

	TObject valueObj;
	setclvalue(&valueObj, cl);
	SetTableHelper(funcName, valueObj);
}

/**
	Assuming the current object is a table, registers all of the functions
	in [functionList] with the table.

	@param functionList A NULL terminated list of functions to register in
		the current table object.
**/
void LuaObject::Register( const LuaFunctionList* functionList )
{
	assert(m_state);
	int i = 0;
	while ( true )
	{
		if ( !functionList[ i ].name )
			break;
		Register( functionList[ i ].name, functionList[ i ].luaStateCFunction );
		i++;
	}
}

/**
	Assuming the current object is a table, registers all of the functions
	in [functionList] with the table.

	@param functionList A NULL terminated list of functions to register in
		the current table object.
**/
void LuaObject::Register( const LuaFunctorList* functionList )
{
	assert(m_state);
	int i = 0;
	while ( true )
	{
		if ( !functionList[ i ].name )
			break;
		Register( functionList[ i ].name, functionList[ i ].luaFunctor );
		i++;
	}
}

/**
	Assuming the current object is a table, unregisters the function called
	[funcName].

	@param funcName The name of the function to unregister.
**/
void LuaObject::Unregister( const char* funcName )
{
	assert(m_state);
	SetNil(funcName);
}

/**
	Assuming the current object is a table, unregisters the list of
	functions described in [functionList].

	@param functionList The list of functions to unregister.
**/
void LuaObject::Unregister( const LuaFunctionList* functionList )
{
	assert(m_state);
	int i = 0;
	while ( true )
	{
		if ( !functionList[ i ].name )
			break;
		Unregister( functionList[ i ].name );
		i++;
	}
}

/**
	Assuming the current object is a table, unregisters the list of
	functions described in [functionList].

	@param functionList The list of functions to unregister.
**/
void LuaObject::Unregister( const LuaFunctorList* functionList )
{
	assert(m_state);
	int i = 0;
	while ( true )
	{
		if (!functionList[ i ].name)
			break;
		Unregister( functionList[i].name );
		i++;
	}
}


inline void LuaObject::SetTableHelper(const char* key, TObject& valueObj)
{
	assert(m_state);
	lua_State* L = m_state->m_state;

	TObject keyObj;
	setsvalue(&keyObj, luaS_newlstr(L, key, strlen(key)));

	luaV_settable(L, &m_object, &keyObj, &valueObj);
}


inline void LuaObject::SetTableHelper(int key, TObject& valueObj)
{
	assert(m_state);
	lua_State* L = m_state->m_state;

	TObject keyObj;
	setnvalue(&keyObj, key);

	luaV_settable(L, &m_object, &keyObj, &valueObj);
}


inline void LuaObject::SetTableHelper(LuaObject& key, TObject& valueObj)
{
	assert(m_state);
	lua_State* L = m_state->m_state;
	luaV_settable(L, &m_object, &key.m_object, &valueObj);
}

} // namespace LuaPlus
