///////////////////////////////////////////////////////////////////////////////
// This source file is part of the LuaPlus source distribution and is Copyright
// 2002 by Joshua C. Jensen (jjensen@workspacewhiz.com).
//
// The latest version may be obtained from http://wwhiz.com/LuaPlus/.
//
// The code presented in this file may be freely used and modified for all
// non-commercial and commercial purposes.
///////////////////////////////////////////////////////////////////////////////
#pragma once

#include <stdio.h>
#include <malloc.h>

///////////////////////////////////////////////////////////////////////////////
// namespace LuaPlus
///////////////////////////////////////////////////////////////////////////////
namespace LuaPlus
{

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Output file helper class.
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
/**
	The DumpObject() facility uses a LuaStateOutFile derived class to
	output data to.  The LuaStateOutFile class may be derived from to enable
	an application specific method of output.
**/
class LUA_API LuaStateOutFile
{
public:
	LuaStateOutFile() : m_file( NULL ), m_fileOwner( false ) {}
	virtual ~LuaStateOutFile()
	{
		if ( m_file  &&  m_fileOwner )
			fclose( m_file );
	}

	virtual bool Open( const char* fileName )
	{
		Close();

		m_file = fopen( fileName, "wb" );
		m_fileOwner = true;

		return m_file != NULL;
	}

	virtual void Close()
	{
		if ( m_file  &&  m_fileOwner )
			fclose( m_file );
	}

	virtual void Print( const char* str, ... )
	{
		char message[ 800 ];
		va_list arglist;

		va_start( arglist, str );
		vsprintf( message, str, arglist );
		va_end( arglist );

		fputs( message, m_file );
	}

	bool Assign( FILE* file )
	{
		m_file = file;
		m_fileOwner = false;

		return true;
	}
	
	void Indent( unsigned int indentLevel );

protected:
	FILE* m_file;
	bool m_fileOwner;
};


class LUA_API LuaStateOutString
    : public LuaStateOutFile
{
public:
    LuaStateOutString() {buffSize = 50; buff = (char*)malloc(buffSize);buff[0]=0;}
    ~LuaStateOutString() {free(buff);}

	virtual bool Open( const char* fileName )
	{
        buff[0] = 0;
		return true;
	}
	virtual void Close()
	{
	}

	virtual void Print( const char* str, ... )
	{
        char * buffStart = buff + strlen(buff);
		va_list arglist;
		va_start( arglist, str );
        while (_vsnprintf(buffStart,buffSize - (buffStart - buff), str, arglist ) == -1) {
            char * old = buff;
            buffSize+=100;
            buff = (char*)malloc(buffSize);
            strcpy(buff , old);
            free(old);
            buffStart += buff - old;
        }
		va_end( arglist );
	}

	bool Assign( FILE* file )
	{
		return false;
	}
	
	void Indent( unsigned int indentLevel );

public:
    const char * GetBuff() {return buff;}

private:
    char * buff;
    size_t buffSize;
};


} // namespace LuaPlus


